#!/usr/bin/env python3
"""
Windows Process Manager (System Management Tool)

Parameters:
    --list: List running processes
    --info PID: Show process details
    --kill PID: Terminate process
    --priority PID LEVEL: Set process priority
    --cpu PID: Show CPU usage
    --memory PID: Show memory usage
    --format FORMAT: Output format (table/json)

Requires: psutil, win32api, tabulate packages and admin privileges
Warning: Terminating system processes may cause system instability
"""

import os
import sys
import psutil
import time
import json
import argparse
import logging
from typing import List, Dict, Any, Optional
from tabulate import tabulate
import win32security
import win32api
import win32process
import win32con
from datetime import datetime

class ProcessManager:
    def __init__(self):
        self.logger = logging.getLogger(__name__)
        self.priority_map = {
            win32process.IDLE_PRIORITY_CLASS: "Idle",
            win32process.BELOW_NORMAL_PRIORITY_CLASS: "Below Normal",
            win32process.NORMAL_PRIORITY_CLASS: "Normal",
            win32process.ABOVE_NORMAL_PRIORITY_CLASS: "Above Normal",
            win32process.HIGH_PRIORITY_CLASS: "High",
            win32process.REALTIME_PRIORITY_CLASS: "Realtime"
        }

    def get_process_list(self) -> List[Dict[str, Any]]:
        """Get list of all running processes with details."""
        processes = []
        for proc in psutil.process_iter(['pid', 'name', 'cpu_percent', 
                                       'memory_percent', 'num_threads', 
                                       'username', 'status']):
            try:
                process_info = proc.info
                process_info['cpu_percent'] = proc.cpu_percent(interval=0.1)
                process_info['memory_mb'] = proc.memory_info().rss / (1024 * 1024)
                process_info['priority'] = self.get_process_priority(proc.pid)
                process_info['start_time'] = datetime.fromtimestamp(
                    proc.create_time()
                ).strftime('%Y-%m-%d %H:%M:%S')
                processes.append(process_info)
            except (psutil.NoSuchProcess, psutil.AccessDenied, 
                    psutil.ZombieProcess) as e:
                self.logger.debug(f"Error getting process info: {str(e)}")
                continue
        return processes

    def get_process_priority(self, pid: int) -> str:
        """Get process priority level."""
        try:
            handle = win32api.OpenProcess(win32con.PROCESS_QUERY_INFORMATION, 
                                        False, pid)
            priority = win32process.GetPriorityClass(handle)
            return self.priority_map.get(priority, "Unknown")
        except Exception as e:
            self.logger.debug(f"Error getting priority for PID {pid}: {str(e)}")
            return "Unknown"

    def set_process_priority(self, pid: int, priority: str) -> bool:
        """Set process priority level."""
        priority_values = {
            "idle": win32process.IDLE_PRIORITY_CLASS,
            "below_normal": win32process.BELOW_NORMAL_PRIORITY_CLASS,
            "normal": win32process.NORMAL_PRIORITY_CLASS,
            "above_normal": win32process.ABOVE_NORMAL_PRIORITY_CLASS,
            "high": win32process.HIGH_PRIORITY_CLASS,
            "realtime": win32process.REALTIME_PRIORITY_CLASS
        }
        
        if priority.lower() not in priority_values:
            self.logger.error(f"Invalid priority level: {priority}")
            return False
            
        try:
            handle = win32api.OpenProcess(win32con.PROCESS_SET_INFORMATION, 
                                        False, pid)
            win32process.SetPriorityClass(handle, priority_values[priority.lower()])
            return True
        except Exception as e:
            self.logger.error(f"Error setting priority for PID {pid}: {str(e)}")
            return False

    def kill_process(self, pid: int) -> bool:
        """Terminate a process by PID."""
        try:
            process = psutil.Process(pid)
            process.terminate()
            return True
        except Exception as e:
            self.logger.error(f"Error terminating process {pid}: {str(e)}")
            return False

    def get_process_tree(self, pid: Optional[int] = None) -> Dict[str, Any]:
        """Get process tree starting from given PID or system processes."""
        def build_tree(parent_pid: int, seen: set = None) -> Dict[str, Any]:
            if seen is None:
                seen = set()
            if parent_pid in seen:
                return None  # Prevent circular references
            seen.add(parent_pid)
            
            try:
                parent = psutil.Process(parent_pid)
                children = []
                for child in parent.children():
                    child_tree = build_tree(child.pid, seen)
                    if child_tree:
                        children.append(child_tree)
                
                return {
                    'pid': parent_pid,
                    'name': parent.name(),
                    'cpu_percent': parent.cpu_percent(interval=0.1),
                    'memory_mb': parent.memory_info().rss / (1024 * 1024),
                    'children': children
                }
            except (psutil.NoSuchProcess, psutil.AccessDenied, 
                    psutil.ZombieProcess):
                return None
        
        if pid:
            return build_tree(pid)
        else:
            # Get all top-level processes
            trees = []
            for proc in psutil.process_iter(['pid', 'ppid']):
                try:
                    if proc.ppid() == 0:  # Top-level process
                        tree = build_tree(proc.pid)
                        if tree:
                            trees.append(tree)
                except (psutil.NoSuchProcess, psutil.AccessDenied, 
                        psutil.ZombieProcess):
                    continue
            return {'processes': trees}

    def display_results(self, processes: List[Dict[str, Any]], 
                       output_format: str = 'table',
                       sort_by: str = 'cpu_percent',
                       limit: int = None):
        """Display process information in specified format."""
        # Sort processes
        processes.sort(key=lambda x: x.get(sort_by, 0), reverse=True)
        if limit:
            processes = processes[:limit]
            
        if output_format == 'json':
            print(json.dumps(processes, indent=4, default=str))
        else:
            headers = ['PID', 'Name', 'CPU %', 'Memory (MB)', 'Threads', 
                      'Priority', 'Status', 'User', 'Start Time']
            table_data = [
                [
                    p['pid'],
                    p['name'],
                    f"{p['cpu_percent']:.1f}",
                    f"{p['memory_mb']:.1f}",
                    p['num_threads'],
                    p['priority'],
                    p['status'],
                    p.get('username', 'N/A'),
                    p['start_time']
                ] for p in processes
            ]
            print(tabulate(table_data, headers=headers, tablefmt='grid'))

    def display_process_tree(self, tree: Dict[str, Any], 
                           indent: int = 0,
                           output_format: str = 'text'):
        """Display process tree in specified format."""
        def format_process(proc: Dict[str, Any], level: int = 0) -> str:
            prefix = "  " * level
            return (f"{prefix}[{proc['pid']}] {proc['name']} "
                   f"(CPU: {proc['cpu_percent']:.1f}%, "
                   f"Memory: {proc['memory_mb']:.1f} MB)")
            
        if output_format == 'json':
            print(json.dumps(tree, indent=4))
        else:
            if 'processes' in tree:  # Multiple top-level processes
                for proc in tree['processes']:
                    self._print_tree(proc)
            else:  # Single process tree
                self._print_tree(tree)
                
    def _print_tree(self, node: Dict[str, Any], level: int = 0):
        """Helper function to print process tree."""
        prefix = "  " * level
        print(f"{prefix}[{node['pid']}] {node['name']} "
              f"(CPU: {node['cpu_percent']:.1f}%, "
              f"Memory: {node['memory_mb']:.1f} MB)")
        for child in node.get('children', []):
            self._print_tree(child, level + 1)

def main():
    parser = argparse.ArgumentParser(
        description='Process Manager - Windows process management utility')
    parser.add_argument('--list', action='store_true',
                       help='List all running processes')
    parser.add_argument('--tree', type=int, nargs='?', const=0, metavar='PID',
                       help='Show process tree (optionally from specific PID)')
    parser.add_argument('--kill', type=int, metavar='PID',
                       help='Terminate a process')
    parser.add_argument('--priority', nargs=2, metavar=('PID', 'LEVEL'),
                       help='Set process priority (idle, below_normal, normal, '
                            'above_normal, high, realtime)')
    parser.add_argument('--sort', choices=['cpu_percent', 'memory_mb', 
                                         'num_threads'],
                       default='cpu_percent',
                       help='Sort processes by field')
    parser.add_argument('--format', choices=['table', 'json'], default='table',
                       help='Output format')
    parser.add_argument('--limit', type=int,
                       help='Limit number of processes shown')
    
    args = parser.parse_args()
    
    # Setup logging
    logging.basicConfig(
        level=logging.INFO,
        format='%(asctime)s - %(levelname)s - %(message)s'
    )
    
    pm = ProcessManager()
    
    try:
        if args.kill:
            if pm.kill_process(args.kill):
                print(f"Successfully terminated process {args.kill}")
            else:
                print(f"Failed to terminate process {args.kill}")
                return 1
                
        elif args.priority:
            pid, level = int(args.priority[0]), args.priority[1]
            if pm.set_process_priority(pid, level):
                print(f"Successfully set priority for process {pid}")
            else:
                print(f"Failed to set priority for process {pid}")
                return 1
                
        elif args.tree is not None:
            tree = pm.get_process_tree(args.tree if args.tree != 0 else None)
            pm.display_process_tree(tree, output_format=args.format)
            
        else:  # Default to --list if no other action specified
            processes = pm.get_process_list()
            pm.display_results(processes, args.format, args.sort, args.limit)
            
    except Exception as e:
        logging.error(f"An error occurred: {str(e)}")
        return 1
        
    return 0

if __name__ == "__main__":
    exit(main())
